# import lexer
import time
import sys
from treelib import Node
from lexer import guteLex
from gparser import guteParse
from semantic_analysis import analyze
from utilities import STDWARN, STDOUT, STDERR, dividePrograms, removeComments, printTree, getInorderList


def main():
    start_time = time.time()
    try:
        # Open the file and read all its content
        with open(sys.argv[1], 'r') as f:
            input_ = f.read()
    except Exception as e:
            error = f'File Error: {e}'
            error = error.replace('[Errno 2] ', '')
            error = error.replace(' or directory', '')
            STDERR(error)
            exit()

    # Split the programs up by EOP
    input_ = removeComments(input_)
    programs = dividePrograms(input_)

    [guteCompile(i, program) for i, program in enumerate(programs)]
    # lexer(programs)

    STDWARN(f'--- {time.time() - start_time} seconds ---')


def guteCompile(i, program):
    print('')
    STDWARN(f'------PROGRAM {i+1}------')

    tokens = guteLex(program)
    if not tokens:
        return

    [print(f'{token.type_}......line {token.line_num}') for token in tokens]
    STDOUT(f'LEXED PROGRAM SUCCESSFULLY\n')

    cst, parse_error = guteParse(tokens, i)
    if parse_error:
        return
    STDWARN('CST:')
    printTree(cst)
    STDOUT(f'PARSED PROGRAM SUCCESSFULLY\n')

    # [print(f'{cst.level(cst.get_node(i))-1*"  "}{cst.get_node(i).tag}') for i in cst.expand_tree()]
    cst_nodes = getInorderList(cst, cst.get_node(cst.root), [])
    # for n in cst_nodes:
    #     print(n.tag)

    analyze(cst_nodes, i)


if __name__ == '__main__':
    main()



from treelib import Node, Tree
from tokens import tokenKinds, tokenKindLiteral
from utilities import STDERR, STDOUT, STDWARN


def guteParse(token_stream, program_num):
    STDOUT(f'PARSER:')
    token_kinds = tokenKinds()
    cst = Tree()
    parse_error = False
    curr_node = cst.create_node('<Program>')
    curr_token = 0

    def parseProgram(program_num):
        nonlocal cst, curr_node
        parseBlock()
        matchConsume('T_eof')
        endChildren()

    def parseBlock():
        nonlocal cst, curr_node
        _print('Block')
        curr_node = cst.create_node('<Block>', parent=curr_node.identifier)

        matchConsume('T_l_brace')
        parseStatementList()
        matchConsume('T_r_brace')
        endChildren()

    def parseStatementList():
        nonlocal cst, curr_node
        _print('Statement List')

        curr_node = cst.create_node(
            '<Statement List>', parent=curr_node.identifier)

        if match('T_k_if') or match('T_k_while') or match('T_k_print') or match('T_id') or match('T_k_int') or match('T_k_boolean') or match('T_k_string') or match('T_l_brace'):
            parseStatement()
            parseStatementList()
        else:
            pass
        endChildren()


    def parseStatement():
        nonlocal cst, curr_node, token_stream
        _print('Statement')
        curr_node = cst.create_node('<Statement>', parent=curr_node.identifier)

        if match('T_k_print'):
            parsePrint()

        elif match('T_id'):
            parseAssignment()

        elif match('T_k_int') or match('T_k_boolean') or match('T_k_string'):
            parseVarDecl()

        elif match('T_k_while'):
            parseWhile()

        elif match('T_k_if'):
            parseIf()

        elif match('T_l_brace'):
            parseBlock()

        else:
            parseError(
                f'Expected if, while, print, ID, int, boolean, string or block')

        endChildren()

    def parsePrint():
        nonlocal cst, curr_node
        _print('Print Statement')

        curr_node = cst.create_node(
            '<Print Statement>', parent=curr_node.identifier)
        matchConsume('T_k_print')
        matchConsume('T_l_paren')
        parseExpr()
        matchConsume('T_r_paren')
        endChildren()

    def parseAssignment():
        nonlocal cst, curr_node
        _print('Assignment Statement')

        curr_node = cst.create_node(
            '<Assignment Statement>', parent=curr_node.identifier)
        parseID()
        matchConsume('T_assign')
        parseExpr()
        endChildren()

    def parseVarDecl():
        nonlocal cst, curr_node
        print('Variable Decleration')

        curr_node = cst.create_node(
            '<Variable Decleration>', parent=curr_node.identifier)
        parseType()
        parseID()
        endChildren()

    def parseWhile():
        nonlocal cst, curr_node
        _print('While Statement')

        curr_node = cst.create_node(
            '<While Statement>', parent=curr_node.identifier)
        matchConsume('T_k_while')
        parseBooleanExpr()
        parseBlock()
        endChildren()

    def parseIf():
        nonlocal cst, curr_node
        print('If Statement')

        curr_node = cst.create_node(
            '<If Statement>', parent=curr_node.identifier)
        matchConsume('T_k_if')

        parseBooleanExpr()
        parseBlock()
        endChildren()

    def parseExpr():
        nonlocal cst, curr_node
        _print('Expression')

        curr_node = cst.create_node('<Expr>', parent=curr_node.identifier)
        if match('T_digit'):
            parseIntExpr()
        elif match('T_l_paren'):
            parseBooleanExpr()
        elif match('T_k_true') or match('T_k_false'):
            parseBooleanExpr()
        elif match('T_quote'):
            parseStringExpr()
        elif match('T_id'):
            parseID()
        else:
            parseError('Expected Int, Bool, String Expressions or an ID')

        endChildren()

    def parseIntExpr():
        nonlocal cst, curr_node
        _print('Int Expression')

        curr_node = cst.create_node('<Int Expr>', parent=curr_node.identifier)
        if match('T_digit'):
            parseDigit()

        if match('T_intop_add'):
            parseIntOp()
            parseExpr()

        endChildren()

    def parseStringExpr():
        nonlocal cst, curr_node
        _print('String Expression')

        curr_node = cst.create_node('<String Expr>', parent=curr_node.identifier)
        matchConsume('T_quote')

        curr_node = cst.create_node('<Char List>', parent=curr_node.identifier)
        _print('Char List')
        parseCharList()
        endChildren()

        matchConsume('T_quote')
        endChildren()

    def parseBooleanExpr():
        nonlocal cst, curr_node
        _print('Boolean Expression')

        curr_node = cst.create_node('<Boolean Expr>', parent=curr_node.identifier)
        if match('T_l_paren'):
            matchConsume('T_l_paren')
            parseExpr()
            parseBoolOp()
            parseExpr()
            matchConsume('T_r_paren')

        elif match('T_k_true') or match('T_k_false'):
            parseBoolVal()

        else:
            parseError('Expected "(", true, or false')

        endChildren()

    def parseID():
        nonlocal cst, curr_node
        _print('ID')

        curr_node = cst.create_node('<ID>', parent=curr_node.identifier)
        if match('T_id'):
            matchConsume('T_id')
        else:
            parseError(f'Expected {tokenKindLiteral("T_id")}')

        endChildren()

    def parseCharList():
        nonlocal cst, curr_node

        if match('T_char'):
            parseChar()
            parseCharList()
        else:
            pass

    def parseType():
        nonlocal cst, curr_node
        _print('Type')
        curr_node = cst.create_node('<Type>', parent=curr_node.identifier)

        if match('T_k_int'):
            matchConsume('T_k_int')
        elif match('T_k_string'):
            matchConsume('T_k_string')
        elif matchConsume('T_k_boolean'):
            matchConsume('T_k_boolean')
        else:
            pass
        endChildren()

    def parseChar():
        nonlocal cst, curr_node
        _print('Character')

        curr_node = cst.create_node('<Char>', parent=curr_node.identifier)

        if matchConsume('T_char'):
            pass
        else:
            pass

        endChildren()

    def parseDigit():
        nonlocal curr_node
        curr_node = cst.create_node('<Digit>', parent=curr_node.identifier)
        _print('Digit')
        if matchConsume('T_digit'):
            return
        else:
            pass

        endChildren()

    def parseBoolOp():
        nonlocal cst, curr_node
        _print('Boolean Operation')

        curr_node = cst.create_node('<Bool Op>', parent=curr_node.identifier)

        if match('T_boolop_eq'):
            matchConsume('T_boolop_eq')
        elif match('T_boolop_ineq'):
            matchConsume('T_boolop_ineq')
        else:
            parseError(f'Expected == or !=')

        endChildren()

    def parseBoolVal():
        nonlocal cst, curr_node
        _print('Boolean Value')
        curr_node = cst.create_node(
            '<BooleanVal>', parent=curr_node.identifier)
        if match('T_k_true'):
            matchConsume('T_k_true')
        if match('T_k_false'):
            matchConsume('T_k_false')
        else:
            pass
        endChildren()

    def parseIntOp():
        nonlocal cst, curr_node
        _print('Integer Operation')
        curr_node = cst.create_node('<Int Op>', parent=curr_node.identifier)
        if match('T_intop_add'):
            matchConsume('T_intop_add')
        else:
            pass
        endChildren()

    def match(to_match):
        nonlocal cst, curr_token, token_kinds, token_stream, curr_node

        if token_stream[curr_token].type_ == to_match and to_match in token_kinds:
            return True

        return False

    def matchConsume(to_match):
        nonlocal cst, curr_token, token_kinds, token_stream, curr_node

        if token_stream[curr_token].type_ == to_match and to_match in token_kinds:
            cst.create_node(f'[ {token_stream[curr_token]} ]',parent=curr_node.identifier, data=token_stream[curr_token])
            curr_token += 1
        else:
            parseError(f'Expected "{tokenKindLiteral(to_match)}"')

    def endChildren():
        nonlocal curr_node

        if cst.get_node(curr_node.bpointer):
            curr_node = cst.parent(curr_node.identifier)

    def parseError(error_message):
        nonlocal curr_token, token_stream, parse_error
        STDERR(
            f'Parse Error: {error_message} | Found "{token_stream[curr_token]}" on line {token_stream[curr_token].line_num}')
        parse_error = True

    def _print(message):
        nonlocal parse_error
        if not parse_error:
            print(message)

    parseProgram(program_num)
    return cst, parse_error



from tokens import *
from utilities import *
import re

def guteLex(program):
    STDOUT(f'LEXER:')

    buffer = []
    tokens = []
    matches = []
    # curr_col = 1
    curr_line = 1
    # comment_len = 0
    string_mode = False
    comment_mode = False
    # ignored_comment = False
    need_newline = False
    newline = '\n'
    whitespace = ' '
    string_quote = '"'
    comment_block = set(('/*', '*/'))
    lookahead = set(('=','!'))
    valid_tokens, token_kinds = grammarDict()
    tokenize = set(('$','"','{','}','(',')','=','==','!='))
    program = program.translate({ord('\t'): '  ', ord('\r'): '\n'})

    def parseError(error_message):
        STDERR(f'Lexer Error: {error_message} | Line: {curr_line}')

    
    program = [*program]

    while program or buffer:
        if not program:
            if not matches:
                parseError(f'Invalid Character {repr(buffer[0])}')
                return
                
            token = spawnToken(matches, token_kinds, valid_tokens,curr_line)
            tokens.append(token)
           

            t_len = len(tokens[-1].value)
            if (buffer[0] == tokens[-1].value):
                del buffer[0]
            else:
                del buffer[:t_len]

            program = program + buffer
            buffer = []
            matches = []

            continue

        next_char = program.pop(0)
        next_next_char = program[0] if 0 < len(program) else None

        if f'{next_char}{next_next_char}' in comment_block:
            comment_mode = not comment_mode
            del program[0]
            continue 

        if comment_mode:
            if next_char == newline:
                curr_line += 1
            continue

        if next_char == whitespace and not string_mode:
            if not buffer:
                continue
            
            if not matches:
                parseError(f'Invalid Character {repr(buffer[0])}')
                return

            token = spawnToken(matches, token_kinds, valid_tokens, curr_line)
            tokens.append(token)

            t_len = len(tokens[-1].value)
            if (buffer[0] == tokens[-1].value):
                del buffer[0]
            else:
                del buffer[:t_len]

            program = buffer + program
            buffer = []
            matches = []

            continue

        if next_char == newline and not string_mode:
            if not buffer:
                curr_line += 1
                continue

            if not matches:
                parseError(f'Invalid Character {repr(buffer[0])}')
                return

            token = spawnToken(matches, token_kinds, valid_tokens, curr_line)
            tokens.append(token)

            curr_line += 1


            t_len = len(tokens[-1].value)
            if (buffer[0] == tokens[-1].value):
                del buffer[0]
            else:
                del buffer[:t_len]

            program = buffer + program
            buffer = []
            matches = []
            continue  

        if next_char in lookahead:
            if f'{next_char}{next_next_char}' in tokenize:
                next_char = f'{next_char}{program.pop(0)}'
                next_next_char = program[0] if 0 < len(program) else None

        if buffer and next_char in tokenize:
            # Make a token
            if not matches:
                parseError(f'Invalid Character {repr(buffer[0])}')
                return

            token = spawnToken(matches, token_kinds,valid_tokens, curr_line)
            tokens.append(token)

            t_len = len(tokens[-1].value)

            if (buffer[0] == tokens[-1].value):
                del buffer[0]
            else:
                del buffer[:t_len]

            program = buffer + [next_char] + program
            buffer = []
            matches = []
            continue
                    
        buffer.append(next_char)

        match = matchToken(''.join(buffer), matches, token_kinds, string_mode)

        if buffer and buffer[-1] == string_quote:
            string_mode = not string_mode
        if match:
            matches.append(match)

    if string_mode:
        parseError('Unterminated string')
        return
    if comment_mode:
        parseError('Unterminated comment')

    return tokens







import re
from beautifultable import BeautifulTable
from operator import itemgetter
from tokens import *
from lexer import *
import click

# Given an input string of programs, split it by the eof symbol ($) and return them as a list


def dividePrograms(pgms):
    input_ = pgms.rstrip()
    eop_count = pgms.count('$')
    programs = pgms.split('$')

    if not programs[0]:
        if len(programs) > 1:
            return ['$' for i in range(eop_count)]

        return []
    else:
        # Filter empty elements in the list
        program_split = tuple(filter(None, programs))

        # Add back the eof character after the splt
        programs = [f'{program_split[i]}$' for i in range(eop_count)]

        # if our program count doesn't match the count of eop's in the input, or the eop count is 0 then we are missing a trailing eop
        program_count = len(programs)
        if eop_count is 0 or eop_count != program_count:
            programs[-1:] = program_split[-1:]
            programs[-1] = programs[-1] + '$'
            STDWARN("Warning: Missing EOP '$'")

        return programs


def removeComments(programs):
    return re.sub(r'(\/\*)(.|\n)+?(\*\/)', '', programs)

# Given a string of chars and a set of valid tokens find and return the kinds of tokens the string matches


def matchToken(char, matches, token_kinds, string_mode):
    for token in token_kinds:
        if string_mode and char != '"' and token != 'T_char':
            continue

        match = re.match(token_kinds[token], char)
        if match:
            matches.append((token, len(char), char))

# Given a list of matches and valid tokens get the longest match or the highest priority match and create a token


def spawnToken(matches, token_kinds, valid_tokens, curr_line):
    # Find Longast match
    longest = max(matches, key=itemgetter(1))
    if type(longest) is tuple:
        longest = [(longest)]

    # Highest priority is always the firstmost element because matches are created in priorty order
    highest_priority = longest[0]
    t_kind = highest_priority[0]
    t_val = highest_priority[2]

    if t_kind == 'T_keyword' or t_kind == 'T_symbol':
        # Instantiate token
        t_kind = valid_tokens[highest_priority[2]][1]

    return Token(t_kind, t_val, curr_line)

# Functions to print to the command line with pretty colors


def STDOUT(message):
    click.echo(click.style(message, fg='green'))


def STDERR(message):
    click.echo(click.style(message, fg='red'))


def STDWARN(message):
    click.echo(click.style(message, fg='yellow'))


def printTree(tree):
    def traverse(t, node):
        print(f'{"-" * t.level(node.identifier)}{node.tag}')
        for node in t.children(node.identifier):
            traverse(t, node)

    traverse(tree, tree.get_node(tree.root))


def getInorderList(tree, node, node_list):
    # def traverse(t, node):
    node_list.append(node)
    for node in tree.children(node.identifier):
        getInorderList(tree, node, node_list)
    return node_list


def traverseInorder(tree, node):
    # def traverse(t, node):
    print(node.tag)
    for node in tree.children(node.identifier):
        traverseInorder(tree, node)

def printSymbolTable(symbol_table):
    table = BeautifulTable()
    table.column_headers = ["Variable", "Type", "Initialized?","Used?","Line Number"]

    # print(symbol_table.ids)
    vars = symbol_table.ids.keys()

    for var in vars:
        details = symbol_table.ids[var]
        table.append_row([var, *details])
   
    print(table)



import re
class Token():
    def __init__(self,type_,value,line_num):
        self.type_ = type_
        self.value = value
        self.line_num = line_num

    def __repr__(self):
        return f'{self.value}'


def grammarDict():
    valid_tokens = {
        # Keywords
        'if': ('(if)(?![\s\S])','T_k_if'),
        'int': ('(int)(?![\s\S])', 'T_k_int'),
        'true': ('(true)(?![\s\S])', 'T_k_true'),
        'false': ('(false)(?![\s\S])', 'T_k_false'),
        'while': ('(while)(?![\s\S])','T_k_while'),
        'print': ('(print)(?![\s\S])','T_k_print'),
        'string': ('(string)(?![\s\S])','T_k_string'),
        'boolean': ('(boolean)(?![\s\S])','T_k_boolean'),

        # Symbols
        '$': ('\$(?![\s\S])', 'T_eof'),
        '"': ('\"(?![\s\S])', 'T_quote'),
        '=': ('=(?![\s\S])', 'T_assign'),
        '{': ('{(?![\s\S])', 'T_l_brace'),
        '}': ('}(?![\s\S])', 'T_r_brace'),
        '(': ('\((?![\s\S])', 'T_l_paren'),
        ')': ('\)(?![\s\S])', 'T_r_paren'),
        '+': ('\+(?![\s\S])', 'T_intop_add'),
        '==': ('==(?![\s\S])', 'T_boolop_eq'),
        '!=': ('!=(?![\s\S])', 'T_boolop_ineq')
    }
    keyword_regex = r'^%s$' % '|'.join(valid_tokens[x][0] for x in valid_tokens)[0:140]
    id_regex = r'^[a-z]$'
    symbol_regex = r'^%s$' % '|'.join(valid_tokens[x][0] for x in valid_tokens)[141:]
    digit_regex = r'^[0-9]$'
    char_regex = r'^[a-z ]\Z'

    t_kinds = {
        'T_keyword': keyword_regex,
        'T_id': id_regex,
        'T_symbol': symbol_regex,
        'T_digit': digit_regex,
        'T_char': char_regex,
    }

    return valid_tokens, t_kinds

def tokenKinds():
    return set((
        # Keywords
        'T_k_if',
        'T_k_int',
        'T_k_true',
        'T_k_false',
        'T_k_while',
        'T_k_print',
        'T_k_string',
        'T_k_boolean',
        'T_eof',
        'T_quote',
        'T_assign',
        'T_l_brace',
        'T_r_brace',
        'T_l_paren',
        'T_r_paren',
        'T_intop_add',
        'T_boolop_eq',
        'T_boolop_ineq',
        'T_id',
        'T_digit',
        'T_char'
    ))

def tokenKindLiteral(token_kind):
    d = {
        'T_k_if':'if',
        'T_k_int':'int',
        'T_k_true':'true',
        'T_k_false': 'false',
        'T_k_while':'while',
        'T_k_print':'print',
        'T_k_string':'string',
        'T_k_boolean':'boolean',
        'T_eof':'$',
        'T_quote':'"',
        'T_assign':'=',
        'T_l_brace':'{',
        'T_r_brace':'}',
        'T_l_paren':'(',
        'T_r_paren':')',
        'T_intop_add':'+',
        'T_boolop_eq':'==',
        'T_boolop_ineq':'!=',
        'T_id':'ID',
        'T_digit':'Digit',
        'T_char': 'Char'
    }
    return d[token_kind]



from treelib import Node, Tree
from scope import gScope
from tokens import Token
from utilities import STDERR, STDOUT, STDWARN, printTree, traverseInorder, printSymbolTable


def makeAST(cst_nodes, program_num):
    STDOUT(f'Semantic Analysis:')
    semantic_error = False
    semantic_warns = []
    symbol_table = []
    curr_scope = None
    curr_node = None
    curr_cst_node = 0
    ast = Tree()


    def parseProgram(program_num):
        nonlocal ast, curr_node

        matchNode(cst_nodes[curr_cst_node].tag, '<Program>')
        parseBlock()
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_eof')
        endChildren()

    def parseBlock():
        nonlocal ast, curr_node, curr_cst_node
        matchNode(cst_nodes[curr_cst_node].tag, "<Block>")
        newScope()

        if ast.size() is 0:
            curr_node = ast.create_node('<Block>')
        else:
            curr_node = ast.create_node('<Block>', parent=curr_node.identifier)

        matchNode(cst_nodes[curr_cst_node].data.type_, "T_l_brace")
        parseStatementList()
        matchNode(cst_nodes[curr_cst_node].data.type_, "T_r_brace")

        if len(symbol_table) > 1:
            endScope()
        endChildren()

    def parseStatementList():
        nonlocal ast, curr_node, curr_cst_node

        matchNode(cst_nodes[curr_cst_node].tag, '<Statement List>')
        if match(cst_nodes[curr_cst_node+1].tag, '<If Statement>') or match(cst_nodes[curr_cst_node+1].tag, '<While Statement>') or match(cst_nodes[curr_cst_node+1].tag, '<Print Statement>') or match(cst_nodes[curr_cst_node+1].tag, '<Assignment Statement>') or match(cst_nodes[curr_cst_node+1].tag, '<Variable Decleration>') or match(cst_nodes[curr_cst_node+1].tag, '<Block>'):
            parseStatement()
            parseStatementList()
        else:

            pass

    def parseStatement():
        matchNode(cst_nodes[curr_cst_node].tag, '<Statement>')

        if match(cst_nodes[curr_cst_node].tag, '<Print Statement>'):
            parsePrint()

        elif match(cst_nodes[curr_cst_node].tag, '<Assignment Statement>'):
            parseAssignment()

        elif match(cst_nodes[curr_cst_node].tag, '<Variable Decleration>'):
            parseVarDecl()

        elif match(cst_nodes[curr_cst_node].tag, '<While Statement>'):
            parseWhile()

        elif match(cst_nodes[curr_cst_node].tag, '<If Statement>'):
            parseIf()

        elif match(cst_nodes[curr_cst_node].tag, '<Block>'):
            parseBlock()

        else:
            semanticError(f'Expected if, while, print, ID, int, boolean, string or block')

    def parsePrint():
        nonlocal ast, curr_node

        curr_node = ast.create_node('<Print Statement>', parent=curr_node.identifier)
        nextCSTNode()
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_k_print')
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_l_paren')
        parseExpr()
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_r_paren')
        endChildren()

    def parseAssignment():
        nonlocal ast, curr_node

        curr_node = ast.create_node('<Assignment Statement>', parent=curr_node.identifier)
        nextCSTNode()

        parseID()
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_assign')
        parseExpr()

        id_info = ast.children(curr_node.identifier)
        var_ = id_info[0].data.value
        value_ = id_info[1]
        line_ = id_info[0].data.line_num

        var_exists = checkID(var_)

        if var_exists:
            
            while not value_.data:
                value_ = ast.children(value_.identifier)[0]

            id_scope = checkScopeLevel(var_)

            var_type = symbol_table[id_scope].checkType(var_)
            value_type = value_.data.type_

            if var_type != value_type:
                scopeError(f'type mismatch, var {var_} cannot equal "{value_.data}"', line_)
                
            checkInit(var_)
        else:
            scopeError(f'variable "{var_}" has not been declared', line_)

        endChildren()

    def parseWhile():
        nonlocal ast, curr_node

        curr_node = ast.create_node('<While Statement>', parent=curr_node.identifier)
        nextCSTNode()
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_k_while')
        parseBooleanExpr()
        parseBlock()
        endChildren()

    def parseIf():
        nonlocal ast, curr_node

        curr_node = ast.create_node('<If Statement>', parent=curr_node.identifier)
        nextCSTNode()
        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_k_if')
        parseBooleanExpr()
        parseBlock()
        endChildren()

    def parseVarDecl():
        nonlocal ast, curr_node, symbol_table
        curr_node = ast.create_node('<Variable Decleration>', parent=curr_node.identifier)
        matchNode(cst_nodes[curr_cst_node].tag, '<Variable Decleration>')
        parseType()
        parseID()

        id_info = ast.children(curr_node.identifier)
        name_ = id_info[1].data.value
        type_ = id_info[0].data.type_
        line_ = id_info[1].data.line_num

        id_exists = checkID(name_)
        if not id_exists:
            _print(f'Added var {name_}:({type_})')
            curr_scope.new_id(name_,type_,line_)
        else:
            scopeError(f'variable "{name_}" already declared', line_)

        endChildren()

    def parseExpr():
        nonlocal ast, curr_node, cst_nodes, curr_cst_node, curr_scope

        matchNode(cst_nodes[curr_cst_node].tag, '<Expr>')

        if match(cst_nodes[curr_cst_node].tag, '<Int Expr>'):
            parseIntExpr()
        elif match(cst_nodes[curr_cst_node].tag, '<Boolean Expr>'):
            parseBooleanExpr()
        elif match(cst_nodes[curr_cst_node].tag, '<String Expr>'):
            nextCSTNode()
            parseStringExpr()
        elif match(cst_nodes[curr_cst_node].tag, '<ID>'):
            parseID()

            var_info = ast.children(curr_node.identifier)
            var_ = var_info[0].data.value


            if checkID(var_):
                id_scope = checkScopeLevel(var_)
                line_ = symbol_table[id_scope].getLineNum(var_)

                scope_used = 0

                if checkInit(var_):
                    scope_used = checkUsed(var_)
                else:
                    scopeWarning(f'variable {var_} has not been initialized',line_)

                symbol_table[scope_used].used(var_)
            else:
                pass

        elif cst_nodes[curr_cst_node].data and match(cst_nodes[curr_cst_node].data.type_, 'T_l_paren'):
            parseBooleanExpr()
        else:
            semanticError('Expected Int, Bool, String Expressions or an ID')

    def parseIntExpr():
        nonlocal ast, curr_node,curr_scope

        matchNode(cst_nodes[curr_cst_node].tag, '<Int Expr>')
        if matchNode(cst_nodes[curr_cst_node].tag, '<Digit>'):
            digit = parseDigit()

            if match(cst_nodes[curr_cst_node].tag, '<Int Op>'):
                nextCSTNode()
                intop = parseIntOp()
                ast.add_node(digit,  parent=curr_node.identifier)
                parseExpr()

                digit_ = ast.children(curr_node.identifier)[0]
                to_add = ast.children(curr_node.identifier)[1]

                while not to_add.data:
                    to_add = ast.children(to_add.identifier)[0]

        
                digit_type = digit_.data.type_
                
                if to_add.data.type_ == 'T_id':
                    id_scope = checkScopeLevel(to_add.data.value)
                    to_add_type = symbol_table[id_scope].checkType(to_add.data.value)
                else:
                    to_add_type = to_add.data.type_

                if digit_type != to_add_type:
                    error_ = ""
                    if to_add.data.type_ == 'T_id':
                        error_ = f'type mismatch, {digit.data} cannot be added to var {to_add.data.value}'
                    else:
                        error_ = f'type mismatch, {digit.data} cannot be added to "{to_add.data.value}"'
                    
                    scopeError(error_, digit_.data.line_num)

                endChildren()
            else:
                ast.add_node(digit,  parent=curr_node.identifier)

    def parseStringExpr():
        nonlocal ast, curr_node

        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_quote')
        curr_node = ast.create_node('<Char List>', parent=curr_node.identifier)
        matchNode(cst_nodes[curr_cst_node].tag, '<Char List>')
        parseCharList()
        endChildren()

        matchNode(cst_nodes[curr_cst_node].data.type_, 'T_quote')
        endChildren()

    def parseBooleanExpr():
        nonlocal ast, curr_node, curr_cst_node

        matchNode(cst_nodes[curr_cst_node].tag, '<Boolean Expr>')
        if cst_nodes[curr_cst_node].data and match(cst_nodes[curr_cst_node].data.type_, 'T_l_paren'):
            nextCSTNode()

            counter = 0
            while not match(cst_nodes[curr_cst_node].tag, '<Bool Op>'):
                counter += 1
                nextCSTNode()

            parseBoolOp()
            curr_cst_node -= counter

            parseExpr()
            parseExpr()


            left = ast.children(curr_node.identifier)[0]
            right = ast.children(curr_node.identifier)[1]

            while not left.data:
                left = ast.children(left.identifier)[0]
            while not right.data:
                right = ast.children(right.identifier)[0]

            _line = left.data.line_num

            if left.data.type_ == 'T_id':
                left_scope = checkScopeLevel(left.data.value)
                left = symbol_table[left_scope].checkType(left.data.value)
            else:
                left = left.data.type_

            if right.data.type_ == 'T_id':
                right_scope = checkScopeLevel(right.data.value)
                right = symbol_table[right_scope].checkType(right.data.value)
            else:
                right = right.data.type_

            if left != right:
                scopeError(f'type mismatch, {left} cannot equal {right}==',_line)

            endChildren()
            matchNode(cst_nodes[curr_cst_node].data.type_, 'T_r_paren')

        elif match(cst_nodes[curr_cst_node].tag, '<BooleanVal>'):
            nextCSTNode()
            bool_node = parseBoolVal()
            ast.add_node(bool_node,  parent=curr_node.identifier)

        else:
            semanticError('Expected "(", true, or false')

    def parseID():
        nonlocal ast, curr_node

        matchNode(cst_nodes[curr_cst_node].tag, '<ID>')
        if match(cst_nodes[curr_cst_node].data.type_, 'T_id'):
            ast.create_node(cst_nodes[curr_cst_node].tag, parent=curr_node.identifier, data=cst_nodes[curr_cst_node].data)
            nextCSTNode()
        else:
            semanticError(f'Expected an ID')

    def parseCharList():
        nonlocal ast, curr_node

        char_list = []
        while match(cst_nodes[curr_cst_node].tag, '<Char>'):
            nextCSTNode()
            match(cst_nodes[curr_cst_node].data.type_, 'T_char')
            char_list.append(cst_nodes[curr_cst_node])
            nextCSTNode()

        if char_list:
            char_list_token = Token('T_k_string', "".join([char.tag[2] for char in char_list]), cst_nodes[curr_cst_node-1].data.line_num)
            curr_node = ast.create_node(f'[ {char_list_token.value} ]', parent=curr_node.identifier, data=char_list_token)

        else:
            char_list_token = Token('T_k_string', ' ', cst_nodes[curr_cst_node-2].data.line_num)
            curr_node = ast.create_node(f'[  ]', parent=curr_node.identifier, data=char_list_token)

    def parseType():
        nonlocal ast, curr_node
        matchNode(cst_nodes[curr_cst_node].tag, '<Type>')

        if match(cst_nodes[curr_cst_node].data.type_, 'T_k_int') or match(cst_nodes[curr_cst_node].data.type_, 'T_k_string') or match(cst_nodes[curr_cst_node].data.type_, 'T_k_boolean'):
            ast.create_node(cst_nodes[curr_cst_node].tag, parent=curr_node.identifier, data=cst_nodes[curr_cst_node].data)
            nextCSTNode()
        else:
            pass

    def parseDigit():
        nonlocal ast, curr_node
        if match(cst_nodes[curr_cst_node].data.type_, 'T_digit'):
            cst_nodes[curr_cst_node].data.type_ = 'T_k_int'
            digit_node = Node(cst_nodes[curr_cst_node].tag, data=cst_nodes[curr_cst_node].data)
            nextCSTNode()
            return digit_node
        else:
            pass

    def parseBoolOp():
        nonlocal ast, curr_node, curr_cst_node, cst_nodes

        matchNode(cst_nodes[curr_cst_node].tag, '<Bool Op>')
        del cst_nodes[curr_cst_node-1]
        curr_cst_node -= 1

        if match(cst_nodes[curr_cst_node].data.type_, 'T_boolop_eq'):
            curr_node = ast.create_node('<IsEqual>', parent=curr_node.identifier)
            nextCSTNode()
            del cst_nodes[curr_cst_node-1]
            curr_cst_node -= 1
        elif match(cst_nodes[curr_cst_node].data.type_, 'T_boolop_ineq'):
            nextCSTNode()
            curr_node = ast.create_node('<NotEqual>', parent=curr_node.identifier)
            del cst_nodes[curr_cst_node-1]
            curr_cst_node -= 1
        else:
            pass

    def parseBoolVal():
        nonlocal ast, curr_node
        bool_node = None

        if match(cst_nodes[curr_cst_node].data.type_, 'T_k_true') or match(cst_nodes[curr_cst_node].data.type_, 'T_k_false'):
            cst_nodes[curr_cst_node].data.type_ = 'T_k_boolean'
            bool_node = Node(cst_nodes[curr_cst_node].tag,data=cst_nodes[curr_cst_node].data)
            nextCSTNode()
        else:
            pass

        if bool_node:
            return bool_node

    def parseIntOp():
        nonlocal ast, curr_node

        if matchNode(cst_nodes[curr_cst_node].data.type_, 'T_intop_add'):
            curr_node = ast.create_node('<Add>', parent=curr_node.identifier)
        else:
            pass

    def match(input_node, expected_node):
        nonlocal curr_cst_node

        if not input_node:
            return False

        if input_node == expected_node:
            return True
        else:
            return False

    def matchNode(input_node, expected_node):
        nonlocal curr_cst_node
        
        if not input_node:
            return False

        if input_node == expected_node:
            nextCSTNode()
            return True
        else:
            semanticError(f'Expected "{expected_node}"')
            return False

    def endChildren():
        nonlocal ast, curr_node

        if ast.get_node(curr_node.bpointer):
            curr_node = ast.parent(curr_node.identifier)

    def nextCSTNode():
        nonlocal curr_cst_node
        curr_cst_node += 1

    def newScope():
        nonlocal symbol_table,curr_scope
        symbol_table.append(gScope())
        curr_scope = symbol_table[-1]
        _print(f'Creating new scope: {symbol_table.index(curr_scope)}')

    def endScope():
        nonlocal symbol_table,curr_scope
        _print(f'Ending scope:{symbol_table.index(curr_scope)}')
        symbol_table.pop()
        curr_scope = symbol_table[-1]

    def checkID(id):
        nonlocal symbol_table

        id_exists = [_scope.getLineNum(id) for _scope in symbol_table[::-1] if _scope.checkExists(id)]

        if id_exists:
            return id_exists[0] 
        return False

    def checkInit(id):
        nonlocal symbol_table, curr_scope

        id_init = [scope.getLineNum(id) for scope in symbol_table[::-1] if scope.checkInit(id)]

        if id_init:
            return id_init[0]
        return False

    def checkUsed(id):
        nonlocal symbol_table, curr_scope

        id_used = [scope.getLineNum(id) for scope in symbol_table[::-1] if scope.checkUsed(id)]

        if id_used:
            return id_used[0]
        return False

    def checkScopeLevel(id):
        nonlocal symbol_table, curr_scope

        for scope in symbol_table[::-1]:
            if scope.checkExists(id):
                return symbol_table.index(scope)
        return None

    def checkFinalWarnings():
        nonlocal curr_scope

        for var, details in curr_scope.ids.items():
            if not curr_scope.checkInit(var):
                line_ = curr_scope.getLineNum(var)
                scopeWarning(f'variable {var} has not been initialized', line_)
            if not curr_scope.checkUsed(var):
                line_ = curr_scope.getLineNum(var)
                scopeWarning(f'variable {var} has not been used', line_)

        pass
    
    def _print(message):
        nonlocal semantic_error
        if not semantic_error:
            print(message)
    
    def semanticError(error_message):
        nonlocal cst_nodes, curr_cst_node, semantic_error
        if not semanticError:
            STDERR(f'Semantic Error: {error_message} | Found "{cst_nodes[curr_cst_node].tag}"')
            semantic_error = True
    
    def scopeError(error_message, line_num):
        nonlocal cst_nodes, curr_cst_node, semantic_error
        
        if not semantic_error:
            STDERR(f'Semantic Error: {error_message} at line {line_num}')
            semantic_error = True

    def scopeWarning(warn_message, line_num):
        nonlocal cst_nodes, curr_cst_node, semantic_warns
        semantic_warns.append(f'Semantic Warning: {warn_message} at line {line_num}')
        semantic_error = True
        
    parseProgram(program_num)

    checkFinalWarnings()

    return ast, symbol_table, semantic_error, semantic_warns




def analyze(cst_nodes, program_num):
    ast, symbol_table, semantic_errors, semantic_warnings = makeAST(cst_nodes, program_num)

    if semantic_errors:
        return

    STDWARN(f'\nProgram {program_num} AST:')
    printTree(ast)
    print()
    if semantic_warnings:
        [STDWARN(warning) for warning in semantic_warnings]
        
    STDWARN(f'\nProgram {program_num} symbol table:')
    printSymbolTable(*symbol_table)



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

# Installs requirements and sets the environment variable
setup(
    name='gutec',
    version='0.3.3',
    author="Michael Gutierrez",
    author_email="mike97.gutierrez@gmail.com",
    description="An Alan++ complient compiler",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=['Click', 'treelib', 'BeautifulTable'],
    py_modules=['./gutec/cli'],
    entry_points='''
        [console_scripts]
        gutec=cli:cli
    ''',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)


class gScope:
    
    def __init__(self):
        self.ids = {}

    # ID attributes: name, type, isInitialized?, isUsed?, line_number
    def new_id(self, name, type_, line_num):
        self.ids[name] = (type_, False, False,line_num)

    def getLineNum(self,name):
        return self.ids[name][3]

    def checkExists(self, name):
        if bool(self.ids):
            return name in self.ids.keys()
        return False

    def checkType(self, name):
        if bool(self.ids):
            return self.ids[name][0]
        return False

    def checkInit(self, name):
        if bool(self.ids):
            return self.ids[name][1]
        return False

    def checkUsed(self, name):
        if bool(self.ids):
            return self.ids[name][2]
        return False

    def init(self, name):
        update_var = (self.ids[name][0], True, self.ids[name][2],  self.ids[name][3])
        del self.ids[name]
        self.ids[name] = update_var

    def used(self, name):
        update_var = (self.ids[name][0], True,self.ids[name][2],  self.ids[name][3])
        del self.ids[name]
        self.ids[name] = update_var

    def getIDs(self):
        return self.ids

    def __repr__(self):
        s = ""
        for k, v in self.ids.items():
            s+= f'{k}:{v}\n'

        return s

    def _print(self):
        print(self.ids.items())
        



import re
def grammarDict():
    valid_tokens = {
        # Keywords
        'T_K_if'      : '(if)(?![\s\S])',
        'T_K_int'     : '(int)(?![\s\S])',
        'T_K_true'    : '(true)(?![\s\S])',
        'T_K_false'   : '(false)(?![\s\S])',
        'T_K_while'   : '(while)(?![\s\S])',
        'T_K_print'   : '(print)(?![\s\S])',
        'T_K_string'  : '(string)(?![\s\S])',
        'T_K_boolean' : '(boolean)(?![\s\S])',

        # Symbols
        'T_EOF'         : '\$(?![\s\S])',
        'T_quote'       : '\"(?![\s\S])',
        'T_var_decl'      : '=(?![\s\S])',
        'T_LBrace'      : '{(?![\s\S])',
        'T_RBrace'      : '}(?![\s\S])',
        'T_LParen'      : '\((?![\s\S])',
        'T_RParen'      : '\)(?![\s\S])',
        'T_intop_add'   : '\+(?![\s\S])',
        'T_boolop_eq'   : '==(?![\s\S])',
        'T_boolop_ineq' : '!=(?![\s\S])',
    }
    keyword_regex = r'^%s$' % '|'.join(valid_tokens[x] for x in valid_tokens)[0:140]
    id_regex = r'^[a-z]$'
    symbol_regex = r'^%s$' % '|'.join(valid_tokens[x] for x in valid_tokens)[141:]
    digit_regex = r'^[0-9]$'
    char_regex = r'^[a-z ]$'

    t_kinds = {
        keyword_regex:('T_keyword',1),
        id_regex:     ('T_ID',     2),
        symbol_regex: ('T_symbol', 3),
        digit_regex:  ('T_digit',  4),
        char_regex:   ('T_char',   5),
    }

    return valid_tokens, t_kinds

    
